package bdd;

import junit.framework.TestCase;

public class BDDTest extends TestCase
{

	BDD bdd = new BDD();
	
	@Override
	protected void setUp() throws Exception 
	{
		bdd = new BDD();
	}

	/**
	 * just run it
	 * filter for covered
	 * nothing much covered: problem integer.max
	 * replace by 7
	 * rerun: heap analysis overflow
	 */
	public void testConstants()
	{
		Function f = new Functions.AndFunction(Functions.TRUE, Functions.TRUE);
		assert(f != null);
		Node node = bdd.build(f);
		assertEquals(Node.ONE, node);
		f = new Functions.AndFunction(Functions.TRUE, Functions.FALSE);
		node = bdd.build(f);
		assertEquals(Node.ZERO, node);
	}
	
	public void testTrue()
	{
		Node n = bdd.build(Functions.TRUE);
		assert(n == Node.ONE);
	}
	
	public void testXor()
	{
		Function x0x1 = new Functions.AndFunction(new Functions.VarFunction(0),
				new Functions.VarFunction(1));
		
		Function x2x3 = new Functions.AndFunction(new Functions.VarFunction(2),
				new Functions.VarFunction(3));
		
		Function f = new Functions.OrFunction(new Functions.AndFunction(x0x1, new Functions.NotFunction(x2x3)),
				new Functions.AndFunction(new Functions.NotFunction(x0x1), x2x3));
		
		assertFalse(f.eval(0, false).eval(1, true).eval(2, false).eval(3, false).getValue());
		
		Node node = bdd.build(f);
//		dump(node);
		assertTrue(node.eval(new boolean[] { true, true, false, false }));
	}
	
	public void testEqualsOrAnd()
	{
		Function x0 = new Functions.VarFunction(0);
		Function x1 = new Functions.VarFunction(1);
		Function x2 = new Functions.VarFunction(2);
		
		// {x0, x2}
		Function x0x2 = new Functions.AndFunction(x0, new Functions.AndFunction(new Functions.NotFunction(x1), x2));
		
		Node x0x2Node = bdd.build(x0x2);
		// {x1}
		Node complNode = bdd.complement(x0x2Node);
		
		assertNotSame(x0x2Node, complNode);
		
		// union
		Node union = bdd.apply(Operators.OR, x0x2Node, complNode);
		assertEquals(Node.ONE, union);
		union = bdd.apply(Operators.OR, complNode, x0x2Node);
		assertEquals(Node.ONE, union);
		
		// intersection
		Node intersection = bdd.apply(Operators.AND, x0x2Node, complNode);
		assertEquals(Node.ZERO, intersection);
		intersection = bdd.apply(Operators.AND, complNode, x0x2Node);
		assertEquals(Node.ZERO, intersection);
	}
	
	public void testComplement()
	{
		Node zero = bdd.complement(Node.ONE);
		assertEquals(Node.ZERO, zero);
		Node one = bdd.complement(Node.ZERO);
		assertEquals(Node.ONE, one);
	}
	
	public void testUnion()
	{
		Function x0 = new Functions.VarFunction(0);
		Function x1 = new Functions.VarFunction(1);
		Function x2 = new Functions.VarFunction(2);
		Function x3 = new Functions.VarFunction(3);
		
		Node n0 = bdd.build(new Functions.AndFunction(x0, new Functions.NotFunction(x1)));
		Node n1 = bdd.build(new Functions.OrFunction(new Functions.NotFunction(x0), x1));
		
		Node union = bdd.apply(Operators.OR, n0, n1);
		assertEquals(Node.ONE, union);
	}
	
	public void testRelProd()
	{
		Node y2 = bdd.make(3, Node.ONE, Node.ZERO);
		Node y1 = bdd.make(2, Node.ZERO, y2);
		Node g = y1;
		Node x2 = bdd.make(1, Node.ZERO, y1);
		y1 = bdd.make(2, Node.ZERO, Node.ONE);
		Node f = bdd.make(0, x2, bdd.make(1, y1, Node.ZERO));
		
		assertTrue(f.eval(new boolean[] { false, true, true, false }));
		assertTrue(f.eval(new boolean[] { true, false, true, true }));
		assertTrue(f.eval(new boolean[] { true, false, true, false }));
		
		// first two array values are dummys
		assertTrue(g.eval(new boolean[] { false, false, true, false }));
		
		Node prod = bdd.relProd(f, g);
		
		assertTrue(prod.eval(new boolean[] { false, true }));
		assertTrue(prod.eval(new boolean[] { true, false }));
		dump(prod);
	}
	
	private static void dump(Node node)
	{
		if (node == null) {
			return;
		}
		
		for (int i = 0; i < indent; i++) {
			System.out.print(" ");
		}
		if (node.getVar() == Integer.MAX_VALUE) {
			System.out.println(node == Node.ONE);
		}
		else {
			System.out.println(node.getVar());
		}
			
		indent += 4;
		dump(node.getLow());
		dump(node.getHigh());
		indent -= 4;
	}
	
	private static int indent = 0;
	
}
